"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
exports.__esModule = true;
exports.options = exports.extensions = void 0;
var postcss_value_parser_1 = __importDefault(require("postcss-value-parser"));
var sass_1 = require("../../utilities/sass");
var types_1 = require("../../utilities/types");
var constants_1 = require("../../utilities/constants");
exports.extensions = constants_1.scss.extensions;
exports.options = __assign({}, constants_1.scss.options);
exports["default"] = (0, sass_1.createSassMigrator)('v9-replace-sass-transition', function (_, _a, context) {
    var methods = _a.methods;
    function insertUnexpectedEasingFunctionComment(node, decl) {
        methods.report({
            severity: 'warning',
            node: decl,
            message: "Unexpected easing function '".concat(node.value, "'. See https://polaris.shopify.com/tokens/motion for possible values.")
        });
    }
    function mutateDurationValue(node, decl) {
        if ((0, sass_1.isPolarisVar)(node)) {
            return;
        }
        if ((0, sass_1.isSassVariable)(node)) {
            methods.report({
                severity: 'warning',
                node: decl,
                message: "Cannot statically analyze SCSS variable ".concat(node.value, ".")
            });
            return;
        }
        var unit = postcss_value_parser_1["default"].unit(node.value);
        if (unit) {
            var constantDuration = "".concat(normalizeStringifiedNumber(unit.number)).concat(unit.unit);
            if (!(0, types_1.isKeyOf)(durationConstantsMap, constantDuration)) {
                methods.report({
                    severity: 'warning',
                    node: decl,
                    message: "No matching duration token for '".concat(constantDuration, "'.")
                });
                return;
            }
            var durationCustomProperty = durationConstantsMap[constantDuration];
            var targetValue = "var(".concat(durationCustomProperty, ")");
            if (context.fix) {
                (0, sass_1.setNodeValue)(node, targetValue);
            }
            else {
                methods.report({
                    severity: 'error',
                    node: decl,
                    message: "Replace duration value with token: ".concat(targetValue)
                });
            }
        }
    }
    function mutateFunctionValue(node, decl) {
        if ((0, sass_1.isPolarisVar)(node)) {
            return;
        }
        if ((0, sass_1.isSassVariable)(node)) {
            methods.report({
                severity: 'warning',
                node: decl,
                message: "Cannot statically analyze SCSS variable ".concat(node.value, ".")
            });
            return;
        }
        if (node.type === 'function') {
            var easingFuncHandlers = {
                // Per the spec, these can all be functions:
                // https://w3c.github.io/csswg-drafts/css-easing/#easing-functions
                linear: insertUnexpectedEasingFunctionComment,
                'cubic-bezier': insertUnexpectedEasingFunctionComment,
                steps: insertUnexpectedEasingFunctionComment
            };
            if ((0, types_1.isKeyOf)(easingFuncHandlers, node.value)) {
                easingFuncHandlers[node.value](node, decl);
                return;
            }
        }
        if (node.type === 'word') {
            if ((0, types_1.isKeyOf)(easingFuncConstantsMap, node.value)) {
                var targetValue = "var(".concat(easingFuncConstantsMap[node.value], ")");
                if (context.fix) {
                    (0, sass_1.setNodeValue)(node, targetValue);
                }
                else {
                    methods.report({
                        severity: 'error',
                        node: decl,
                        message: "Replace easing function with token: ".concat(targetValue)
                    });
                }
                return;
            }
            if (cssEasingBuiltinFuncs.includes(node.value)) {
                insertUnexpectedEasingFunctionComment(node, decl);
            }
        }
    }
    function mutateDelayValue(node, decl) {
        // For now, we treat delays like durations
        return mutateDurationValue(node, decl);
    }
    function mutateShorthandValue(decl, parsedValue) {
        var splitValues = [[]];
        // Gathering up references of nodes into groups. Important to note that
        // we're dealing with mutable structures here, so we are purposefully
        // NOT making copies.
        parsedValue.nodes.forEach(function (node) {
            if (node.type === 'div') {
                splitValues.push([]);
            }
            else {
                splitValues[splitValues.length - 1].push(node);
            }
        });
        splitValues.forEach(function (nodes) {
            // From the spec:
            //
            // Note that order is important within the items in this property: the
            // first value that can be parsed as a time is assigned to the
            // transition-duration, and the second value that can be parsed as a
            // time is assigned to transition-delay.
            // https://w3c.github.io/csswg-drafts/css-transitions-1/#transition-shorthand-property
            //
            // That sounds like an array to me! [0] is duration, [1] is delay.
            var timings = [];
            nodes.forEach(function (node) {
                var unit = postcss_value_parser_1["default"].unit(node.value);
                if ((0, sass_1.isTransformableDuration)(unit)) {
                    timings.push(node);
                }
                else {
                    // This node could be either the property to animate, or an easing
                    // function. We try mutate the easing function, but if not we assume
                    // it's the property to animate and therefore do not leave a comment.
                    mutateFunctionValue(node, decl);
                }
            });
            if (timings[0]) {
                mutateDurationValue(timings[0], decl);
            }
            if (timings[1]) {
                mutateDelayValue(timings[1], decl);
            }
        });
    }
    return function (root) {
        methods.walkDecls(root, function (decl) {
            var handlers = {
                'transition-duration': function () {
                    parsedValue.nodes.forEach(function (node) {
                        mutateDurationValue(node, decl);
                    });
                },
                'transition-delay': function () {
                    parsedValue.nodes.forEach(function (node) {
                        mutateDelayValue(node, decl);
                    });
                },
                'transition-timing-function': function () {
                    parsedValue.nodes.forEach(function (node) {
                        mutateFunctionValue(node, decl);
                    });
                },
                transition: function () {
                    mutateShorthandValue(decl, parsedValue);
                },
                'animation-duration': function () {
                    parsedValue.nodes.forEach(function (node) {
                        mutateDurationValue(node, decl);
                    });
                },
                'animation-delay': function () {
                    parsedValue.nodes.forEach(function (node) {
                        mutateDelayValue(node, decl);
                    });
                },
                'animation-timing-function': function () {
                    parsedValue.nodes.forEach(function (node) {
                        mutateFunctionValue(node, decl);
                    });
                },
                animation: function () {
                    mutateShorthandValue(decl, parsedValue);
                }
            };
            if (!handlers[decl.prop]) {
                return;
            }
            var parsedValue = (0, postcss_value_parser_1["default"])(decl.value);
            if ((0, sass_1.hasNumericOperator)(parsedValue)) {
                methods.report({
                    node: decl,
                    severity: 'warning',
                    message: 'Numeric operator detected.'
                });
            }
            handlers[decl.prop]();
            if (context.fix) {
                decl.value = parsedValue.toString();
            }
        });
    };
});
var durationConstantsMap = {
    '0': '--p-duration-0',
    '0s': '--p-duration-0',
    '0ms': '--p-duration-0',
    '50ms': '--p-duration-50',
    '0.05s': '--p-duration-50',
    '100ms': '--p-duration-100',
    '0.1s': '--p-duration-100',
    '150ms': '--p-duration-150',
    '0.15s': '--p-duration-150',
    '200ms': '--p-duration-200',
    '0.2s': '--p-duration-200',
    '250ms': '--p-duration-250',
    '0.25s': '--p-duration-250',
    '300ms': '--p-duration-300',
    '0.3s': '--p-duration-300',
    '350ms': '--p-duration-350',
    '0.35s': '--p-duration-350',
    '400ms': '--p-duration-400',
    '0.4s': '--p-duration-400',
    '450ms': '--p-duration-450',
    '0.45s': '--p-duration-450',
    '500ms': '--p-duration-500',
    '0.5s': '--p-duration-500',
    '5s': '--p-duration-5000'
};
var easingFuncConstantsMap = {
    linear: '--p-linear',
    ease: '--p-ease',
    'ease-in': '--p-ease-in',
    'ease-out': '--p-ease-out',
    'ease-in-out': '--p-ease-in-out'
};
// Per the spec for transition easing functions:
// https://w3c.github.io/csswg-drafts/css-easing/#easing-functions
var cssEasingBuiltinFuncs = [
    'linear',
    'ease',
    'ease-in',
    'ease-out',
    'ease-in-out',
    'cubic-bezier',
    'step-start',
    'step-end',
    'steps',
];
function normalizeStringifiedNumber(number) {
    return Number(number).toString();
}
